import pandas as pd
import numpy as np
from math import sqrt
from scipy.stats import (norm, linregress )
from . import utils


# portfolio statistics
def total_return(returns):
  """calculates total compounded returns"""
  return returns.add(1).prod() -1

def expected_return(returns, aggregate=None, compounded=True):
  """
  Returns the expected return for a given period
  by calculating the geometric holding period return
  """
  returns = utils.aggregate_returns(returns, aggregate, compounded)
  return np.product(1 + returns) ** (1 / returns.count()) - 1

def avg_return(returns, aggregate=None, compounded=True):
  """Calculates the average return/trade return for a period"""
  if aggregate:
    returns = utils.aggregate_returns(returns, aggregate, compounded)
  return returns[returns != 0].dropna().mean()

def cagr(returns, compounded=True, periods=252):
    """
    Calculates the communicative annualized growth return
    (CAGR%) of access returns
    """
    if compounded:
        total = total_return(returns)
    else:
        total = np.sum(returns)
    years = (returns.count()-1)/periods
    res = abs(total + 1.0) ** (1.0 / years) - 1
    return res

def volatility(returns, periods=252, annualize=True):
  """Calculates the volatility of returns for a period"""
  std = returns.std()
  if annualize:
    return std * np.sqrt(periods)
  return std

def rolling_volatility(returns, rolling_period=126, periods_per_year=252):
  return returns.rolling(rolling_period).std() * np.sqrt(periods_per_year)

def implied_volatility(returns, periods=252, annualize=True):
  """Calculates the implied volatility of returns for a period"""
  logret = utils.to_log_returns(returns)
  if annualize:
    return logret.rolling(periods).std() * np.sqrt(periods)
  return logret.std()

def exponential_stdev(returns, window=30, is_halflife=False):
  """Returns series representing exponential volatility of returns"""
  halflife = window if is_halflife else None
  return returns.ewm(com=None, span=window, halflife=halflife, min_periods=window).std()

def autocorr_penalty(returns):
  """Metric to account for auto correlation"""
  res = {}
  returns = utils.clean(returns)
  for c in returns:
    num = returns[c].count()
    coef = np.abs(np.corrcoef(returns[c][:-1], returns[c][1:])[0, 1])
    corr = [((num - x)/num) * coef ** x for x in range(1, num)]
    res[c] = np.sqrt(1 + 2 * np.sum(corr))
  return pd.DataFrame(res, index=[0]).squeeze(axis=0)

def sharpe(returns, rf, periods=252, annualize=True, smart=False):
  """
  Calculates the sharpe ratio of access returns
  """
  excess_returns = utils.to_excess_returns(returns, rf, periods)
  divisor = excess_returns.std(ddof=1)
  if smart:
    # penalize sharpe with auto correlation
    divisor = divisor * autocorr_penalty(returns)
  res = excess_returns.mean() / divisor
  if annualize:
    return res * np.sqrt(periods)
  return res

def rolling_sharpe(returns, rf, rolling_period=126, annualize=True, periods_per_year=252):
  excess_returns = utils.to_excess_returns(returns, rf, periods_per_year)
  res = excess_returns.rolling(rolling_period).mean() / returns.rolling(rolling_period).std(ddof=1)
  if annualize:
      res = res * np.sqrt(periods_per_year)
  return res

def sortino(returns, rf, periods=252, annualize=True, smart=False):
  """
  Calculates the sortino ratio of access returns
  If rf is non-zero, you must specify periods.
  In this case, rf is assumed to be expressed in yearly (annualized) terms
  Calculation is based on this paper by Red Rock Capital
  http://www.redrockcapital.com/Sortino__A__Sharper__Ratio_Red_Rock_Capital.pdf
  """
  excess_returns = utils.to_excess_returns(returns, rf, periods)
  downside = np.sqrt((excess_returns[excess_returns < 0] ** 2).sum() / len(excess_returns))
  if smart:
    # penalize sortino with auto correlation
    downside = downside * autocorr_penalty(returns)
  res = excess_returns.mean() / downside
  if annualize:
    return res * np.sqrt(periods)
  return res

def rolling_sortino(returns, rf, rolling_period=126, annualize=True, periods_per_year=252 ):
  excess_returns = utils.to_excess_returns(returns, rf, periods_per_year)
  downside = excess_returns.rolling(rolling_period).apply(lambda x: (x.values[x.values < 0]**2).sum()) / rolling_period
  res = excess_returns.rolling(rolling_period).mean() / np.sqrt(downside)
  if annualize:
    res = res * np.sqrt(periods_per_year)
  return res

def adjusted_sortino(returns, rf, periods=252, annualize=True, smart=False):
  """
  Jack Schwager's version of the Sortino ratio allows for
  direct comparisons to the Sharpe. See here for more info:
  https://archive.is/wip/2rwFW
  """
  data = sortino(returns, rf, periods=periods, annualize=annualize, smart=smart)
  return data / sqrt(2)

def rar(returns, rf=0., periods=252):
    """
    Calculates the risk-adjusted return of access returns
    (CAGR / exposure. takes time into account.)
    """
    excess_returns = utils.to_excess_returns(returns, rf, periods)
    return cagr(excess_returns) / utils.exposure(excess_returns)

def omega(returns, required_returns=0., periods=252):
    """
    Determines the Omega ratio of a strategy.
    See https://en.wikipedia.org/wiki/Omega_ratio for more details.
    """
    returns_less_thresh = utils.to_excess_returns(returns, required_returns, periods)      
    numer = returns_less_thresh[returns_less_thresh > 0.0].sum()
    denom = returns_less_thresh[returns_less_thresh < 0.0].sum()
    return numer / abs(denom)

def calmar(returns):
    """Calculates the calmar ratio (CAGR% / MaxDD%)"""
    cagr_ratio = cagr(returns)
    max_dd = max_drawdown(returns)
    return cagr_ratio / abs(max_dd)
  
def gain_to_pain_ratio(returns, resolution="D", periods=252):
    """
    Jack Schwager's GPR. See here for more info:
    https://archive.is/wip/2rwFW
    """
    returns = returns.resample(resolution).sum()
    downside = abs(returns[returns < 0].sum())
    return returns.sum() / downside

def outliers(returns, quantile=.95):
  """Returns series of outliers"""
  return returns[returns > returns.quantile(quantile)].dropna(how='all')

def best(returns, aggregate=None, compounded=True):
  """Returns the best day/month/week/quarter/year's return"""
  return utils.aggregate_returns(returns, aggregate, compounded).max()

def worst(returns, aggregate=None, compounded=True):
  """Returns the worst day/month/week/quarter/year's return"""
  return utils.aggregate_returns(returns, aggregate, compounded).min()

def consecutive_wins(returns, aggregate=None, compounded=True):
  """Returns the maximum consecutive wins by day/month/week/quarter/year"""
  returns = utils.aggregate_returns(returns, aggregate, compounded) > 0
  return utils.count_consecutive(returns).max()

def consecutive_losses(returns, aggregate=None, compounded=True):
  """Returns the maximum consecutive losses by day/month/week/quarter/year"""
  returns = utils.aggregate_returns(returns, aggregate, compounded) < 0
  return utils.count_consecutive(returns).max()

def win_rate(returns, aggregate=None, compounded=True):
  """Calculates the win ratio for a period"""
  if aggregate:
    returns = utils.aggregate_returns(returns, aggregate, compounded)
  return returns[returns > 0].count() / returns[returns != 0].count()

def avg_win(returns, aggregate=None, compounded=True):
  """
  Calculates the average winning
  return/trade return for a period
  """
  if aggregate:
    returns = utils.aggregate_returns(returns, aggregate, compounded)
  return returns[returns > 0].mean()

def avg_loss(returns, aggregate=None, compounded=True):
  """
  Calculates the average low if
  return/trade return for a period
  """
  if aggregate:
    returns = utils.aggregate_returns(returns, aggregate, compounded)
  return returns[returns < 0].mean()

def skew(returns):
    """
    Calculates returns' skewness
    (the degree of asymmetry of a distribution around its mean)
    """
    return returns.skew()

def kurtosis(returns):
    """
    Calculates returns' kurtosis
    (the degree to which a distribution peak compared to a normal distribution)
    """
    return returns.kurtosis()

def max_drawdown(returns):
    """Calculates the maximum drawdown"""
    dd = utils.to_drawdown_series(returns)
    return dd.min()

def avg_drawdown(dd_details):
    """Calculates the maximum drawdown"""
    res = {}
    for c in dd_details.columns.get_level_values(0).unique():
      res[c] = dd_details[c]['max drawdown'].mean()/100
    return pd.DataFrame(res, index=[0]).squeeze(axis=0)
  
def longest_drawdown_days(returns):
    """Calculates the maximum drawdown"""
    dd = utils.to_drawdown_series(returns) < 0
    return utils.count_consecutive(dd).max()

def avg_drawdown_days(returns):
    """Calculates the maximum drawdown"""
    dd = utils.to_drawdown_series(returns) < 0
    return utils.count_consecutive(dd).mean()
  
def ulcer_index(returns):
    """Calculates the ulcer index score (downside risk measurment)"""
    dd = utils.to_drawdown_series(returns)
    return np.sqrt(np.divide((dd**2).sum(), returns.shape[0] - 1))

def ulcer_performance_index(returns, rf=0):
    """
    Calculates the ulcer index score
    (downside risk measurment)
    """
    return total_return(utils.to_excess_returns(returns, rf))/ ulcer_index(returns)

def value_at_risk(returns, sigma=1, confidence=0.95):
  """
  Calculats the daily value-at-risk
  (variance-covariance calculation with confidence n)
  """
  res = {}
  returns = utils.clean(returns)
  for c in returns:
    mu = returns[c].mean()
    sigma = returns[c].std() 
    res[c] = norm.ppf(1-confidence, mu, sigma)
  return pd.DataFrame(res, index=[0]).squeeze(axis=0)

def conditional_value_at_risk(returns, sigma=1, confidence=0.95):
    """
    Calculats the conditional daily value-at-risk (aka expected shortfall)
    quantifies the amount of tail risk an investment
    """
    var = value_at_risk(returns, sigma, confidence)
    c_var = returns[returns < var].mean()
    return c_var

def serenity_index(returns, rf=0):
  """
  Calculates the serenity index score
  (https://www.keyquant.com/Download/GetFile?Filename=%5CPublications%5CKeyQuant_WhitePaper_APT_Part1.pdf)
  """
  dd = utils.to_drawdown_series(returns)
  pitfall = - value_at_risk(dd) / returns.std()
  return ulcer_performance_index(returns, rf) * pitfall

def risk_of_ruin(returns):
    """
    Calculates the risk of ruin
    (the likelihood of losing all one's investment capital)
    """
    wins = win_rate(returns)
    return ((1 - wins) / (1 + wins)) ** returns.count()

def tail_ratio(returns, cutoff=0.95):
    """
    Measures the ratio between the right
    (95%) and left tail (5%).
    """
    return abs(returns.quantile(cutoff) / returns.quantile(1-cutoff))

def payoff_ratio(returns):
    """Measures the payoff ratio (average win/average loss)"""
    return avg_win(returns) / abs(avg_loss(returns))

def profit_ratio(returns):
    """Measures the profit ratio (win ratio / loss ratio)"""
    wins = returns[returns >= 0]
    loss = returns[returns < 0]
    win_ratio = abs(wins.mean() / wins.count())
    loss_ratio = abs(loss.mean() / loss.count())
    return win_ratio / loss_ratio

def profit_factor(returns):
    """Measures the profit ratio (wins/loss)"""
    return abs(returns[returns >= 0].sum() / returns[returns < 0].sum())

def cpc_index(returns):
    """
    Measures the cpc ratio
    (profit factor * win % * win loss ratio)
    """
    return profit_factor(returns) * win_rate(returns) * payoff_ratio(returns)

def common_sense_ratio(returns):
    """Measures the common sense ratio (profit factor * tail ratio)"""
    return profit_factor(returns) * tail_ratio(returns)

def outlier_win_ratio(returns, quantile=.99):
    """
    Calculates the outlier winners ratio
    99th percentile of returns / mean positive return
    """
    return returns.quantile(quantile).mean() / returns[returns >= 0].mean()

def outlier_loss_ratio(returns, quantile=.01):
    """
    Calculates the outlier losers ratio
    1st percentile of returns / mean negative return
    """
    return returns.quantile(quantile).mean() / returns[returns < 0].mean()

def recovery_factor(returns):
  """Measures how fast the strategy recovers from drawdowns"""
  max_dd = max_drawdown(returns)
  return total_return(returns) / abs(max_dd)

def risk_return_ratio(returns):
    """
    Calculates the return / risk ratio
    (sharpe ratio without factoring in the risk-free rate)
    """
    return returns.mean() / returns.std()

def kelly_criterion(returns):
  """
  Calculates the recommended maximum amount of capital that
  should be allocated to the given strategy, based on the
  Kelly Criterion (http://en.wikipedia.org/wiki/Kelly_criterion)
  """
  win_loss_ratio = payoff_ratio(returns)
  win_prob = win_rate(returns)
  lose_prob = 1 - win_prob

  return ((win_loss_ratio * win_prob) - lose_prob) / win_loss_ratio

def r_squared(returns, benchmark):
  returns = utils.clean(returns)
  benchmark = utils.clean(benchmark)
  res = {}
  for c in returns:
    _, _, r_val, _, _ = linregress(returns[c], benchmark.squeeze(axis=1))
    res[c] = r_val**2
  return pd.DataFrame(res, index=[0]).squeeze(axis=0)

def information_ratio(returns, benchmark, periods=1):
  """
  Calculates the information ratio
  (basically the risk return ratio of the net profits)
  """
  
  diff = utils.to_excess_returns(returns, benchmark, periods)
  return diff.mean() / diff.std()

def beta(returns, benchmark ):
  """Calculates beta of the portfolio"""
  # find covariance
  frames = [returns, benchmark]
  df = pd.concat(frames, axis=1)
  matrix = df.cov()
  res = {}
  for c in returns:
    res[c] = matrix[c].iloc[-1]/ matrix.iloc[-1,-1]
  return pd.DataFrame(res, index=[0]).squeeze(axis=0)


def alpha(returns, benchmark, periods=252):
  """Calculates alpha of the portfolio"""
  # find beta
  _beta = beta(returns, benchmark)
  # calculates measures now
  res = {}
  for c in returns:
    res[c] = (returns[c].mean() - _beta[c] * benchmark.mean().values[0] )* periods
  return pd.DataFrame(res, index=[0]).squeeze(axis=0)

def rolling_greeks(returns, benchmark, periods=252):
  """Calculates rolling alpha and beta of the portfolio"""
  frames = [returns, benchmark]
  df = pd.concat(frames, axis=1)
  df.columns = [*df.columns[:-1], 'benchmark']
  df = df.fillna(0)
  corr = df.rolling(int(periods)).corr().unstack()['benchmark'].drop('benchmark', axis=1)
  std = df.rolling(int(periods)).std()
  beta = pd.DataFrame()
  alpha = pd.DataFrame()
  for c in returns:
    beta[c] = corr[c] * std[c] / std['benchmark']
    alpha[c] = df[c].mean() - beta[c] * df['benchmark'].mean()
  return pd.concat([beta,alpha],axis=1,keys=['beta','alpha']).swaplevel(0,1,axis=1).sort_index(axis=1).fillna(0)


def monthly_returns(returns, eoy=True):
    """Calculates monthly returns"""
    if isinstance(returns, pd.DataFrame):
        returns = returns.copy()
        returns.columns = map(str.lower, returns.columns)
        if len(returns.columns) > 1 and 'close' in returns.columns:
            returns = returns['close']
        else:
            returns = returns[returns.columns[0]]

    original_returns = returns.copy()
    returns = pd.DataFrame(
        utils.group_returns(returns,
                            returns.index.strftime('%Y-%m-01')))

    returns.columns = ['Returns']
    returns.index = pd.to_datetime(returns.index)

    # get returnsframe
    returns['Year'] = returns.index.strftime('%Y')
    returns['Month'] = returns.index.strftime('%b')

    # make pivot table
    returns = returns.pivot('Year', 'Month', 'Returns').fillna(0)

    # handle missing months
    for month in ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                  'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']:
        if month not in returns.columns:
            returns.loc[:, month] = 0

    # order columns by month
    returns = returns[['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                       'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']]

    if eoy:
        returns['eoy'] = utils.group_returns(
            original_returns, original_returns.index.year).values

    returns.columns = map(lambda x: str(x).upper(), returns.columns)
    returns.index.name = None

    return returns



import datetime as dt
import calendar
import pandas as pd
import numpy as np
from . import stats
from math import ceil
  
def drawdown_details(drawdown):
  """
  Calculates drawdown details, including start/end/valley dates,
  duration, max drawdown and max dd for 99% of the dd period
  for every drawdown period
  """
  def _drawdown_details(drawdown):
    # mark no drawdown
    no_dd = drawdown == 0

    # extract dd start dates
    starts = ~no_dd & no_dd.shift(1)
    starts = list(starts[starts].index)

    # extract end dates
    ends = no_dd & (~no_dd).shift(1)
    ends = list(ends[ends].index)

    # no drawdown :)
    if not starts:
        return pd.DataFrame(
            index=[], columns=('start', 'valley', 'end', 'days',
                                'max drawdown', '99% max drawdown'))

    # drawdown series begins in a drawdown
    if ends and starts[0] > ends[0]:
        starts.insert(0, drawdown.index[0])

    # series ends in a drawdown fill with last date
    if not ends or starts[-1] > ends[-1]:
        ends.append(drawdown.index[-1])

    # build dataframe from results
    data = []
    for i, _ in enumerate(starts):
        dd = drawdown[starts[i]:ends[i]]
        clean_dd = -remove_outliers(-dd, .99)
        data.append((starts[i], dd.idxmin(), ends[i],
                      (ends[i] - starts[i]).days,
                      dd.min() * 100, clean_dd.min() * 100))

    df = pd.DataFrame(data=data,
                        columns=('start', 'valley', 'end', 'days',
                                'max drawdown',
                                '99% max drawdown'))
    df['days'] = df['days'].astype(int)
    df['max drawdown'] = df['max drawdown'].astype(float)
    df['99% max drawdown'] = df['99% max drawdown'].astype(float)

    df['start'] = df['start'].dt.strftime('%Y-%m-%d')
    df['end'] = df['end'].dt.strftime('%Y-%m-%d')
    df['valley'] = df['valley'].dt.strftime('%Y-%m-%d')

    return df

  if isinstance(drawdown, pd.DataFrame):
    _dfs = {}
    for col in drawdown.columns:
      _dfs[col] = _drawdown_details(drawdown[col])
    return pd.concat(_dfs, axis=1)
  return _drawdown_details(drawdown)

def to_pct (input, dec=1):
  if dec == 2:
    dec = '{:,.2f}%'
  elif dec == 3:
    dec = '{:,.3f}%'
  else:
    dec = '{:,.1f}%'
  return (input*100).map(dec.format)

def to_num (input, dec=1):
  if dec == 2:
    dec = '{:,.2f}'
  elif dec == 3:
    dec = '{:,.3f}'
  elif dec == 0:
    dec = '{:0.0f}'
  else:
    dec = '{:,.1f}'
  return input.map(dec.format)

# date manipulation
# the df.index should be datetime 
# dd/mm/yyyy is the default setting
# begin

def inc_day(date, inc):
  """increment date by some amount of days"""
  return date + dt.timedelta(days=inc)

def inc_month(date, inc):
  """increment date by some amount of months"""
  month = date.month - 1 + inc
  year = date.year + month // 12
  month = month % 12 + 1
  day = min(date.day, calendar.monthrange(year, month)[1])
  return dt.datetime(year, month, day)

def inc_year(date, inc):
  """increment date by some amount of years"""
  year = date.year + inc
  month = date.month
  day = min(date.day, calendar.monthrange(year, month)[1])
  return dt.datetime(year, month, day)

def month_end(date):
  """get last day of month from date given"""
  aux = inc_month(date,1)
  return dt.datetime(aux.year, aux.month , 1) - dt.timedelta(days=1)

def month_beg(date):
  """get first day of month from date given"""
  return dt.datetime(date.year, date.month, 1)

def year_end(date):
  """get last day of year from date given"""
  aux = inc_year(date,1)
  return dt.datetime(aux.year, 1 , 1) - dt.timedelta(days=1)

def year_beg(date):
  """get first day of year from date given"""
  return dt.datetime(date.year, 1 , 1)

def mtd(df, date=None):
  """month to date slice"""
  log_1 = df.index >= month_beg(date)
  log_2 = df.index <= date
  mask = np.logical_and(log_1, log_2)
  return df[mask]

def qtd(df, date):
  """quarter to date slice"""
  log_1 = df.index >= month_beg(inc_month(date, -((date.month -1 ) % 3)))
  log_2 = df.index <= date
  mask = np.logical_and(log_1, log_2)
  return df[mask]

def std(df, date):
  """semester to date slice"""
  log_1 = df.index >= month_beg(inc_month(date, -((date.month -1 ) % 6)))
  log_2 = df.index <= date
  mask = np.logical_and(log_1, log_2)
  return df[mask]

def ytd(df, date):
  """year to date slice"""
  log_1 = df.index >= date.strftime('%Y-01-01')
  log_2 = df.index <= date
  mask = np.logical_and(log_1, log_2)
  return df[mask]

def l3m(df, date):
  """year to date slice"""
  log_1 = df.index >= inc_month(date, -3)
  log_2 = df.index <= date
  mask = np.logical_and(log_1, log_2)
  return df[mask]

def l6m(df, date):
  """year to date slice"""
  log_1 = df.index >= inc_month(date, -6)
  log_2 = df.index <= date
  mask = np.logical_and(log_1, log_2)
  return df[mask]

def l1y(df, date):
  """year to date slice"""
  log_1 = df.index >= inc_year(date, -1)
  log_2 = df.index <= date
  mask = np.logical_and(log_1, log_2)
  return df[mask]

def l3y(df, date):
  """year to date slice"""
  log_1 = df.index >= inc_year(date, -3)
  log_2 = df.index <= date
  mask = np.logical_and(log_1, log_2)
  return df[mask]

def l5y(df, date):
  """year to date slice"""
  log_1 = df.index >= inc_year(date, -5)
  log_2 = df.index <= date
  mask = np.logical_and(log_1, log_2)
  return df[mask]

def l10y(df, date):
  """year to date slice"""
  log_1 = df.index >= inc_year(date, -10)
  log_2 = df.index <= date
  mask = np.logical_and(log_1, log_2)
  return df[mask]

# todo
# 1. calendarios DU brasil DI/B3
# 2. correct datas based on official callendar
# end

# results calculations
# default prices should be given
# default using daily returns
# view price, returns and cota should give the same results

# begin

def clean(data):
  """clean step to remove inconsistend data"""
  return data.copy().fillna(0).replace([np.inf, -np.inf, float('NaN')])

def to_quotes(returns, base=1e5):
  """converts returns series to quote prices"""
  return (1+ to_rolling_returns(returns))* base

def to_returns(prices):
  """simple arithmetic returns from a price series"""
  return prices.pct_change()

def to_rolling_returns(returns):
  """calculates rolling compounded returns"""
  return returns.add(1).cumprod() -1
  
def to_log_returns(returns, base='e'):
  """logarithm returns from a price series and base"""
  if base == '2':
    return np.log2(1+returns)
  elif base == '10':
    return np.log10(1+returns)
  else:
    return np.log(1+returns)

def rebase(prices, base=100):
  """rebase a series to a given initial base"""
  return prices / prices.iloc[0] * base

def to_excess_returns(returns, rf=0., periods=252):
  """Calculates excesss returns"""
  if periods is not None:
    rf = np.power(1 + rf, 1./ periods)  -1.
  if isinstance(returns, pd.DataFrame):
    result = pd.DataFrame()
    for c in returns.columns:
      result[c] = returns[c] - rf
    return result
  return returns - rf

def match_volatility(returns, benchmark): 
  return (returns / returns.std()) * benchmark.std()
  
def group_returns(returns, groupby, compounded=True):
  """Summarize returns
  group_returns(df, df.index.year)
  group_returns(df, [df.index.year, df.index.month])
  """
  if compounded:
      return returns.groupby(groupby).apply(stats.total_return)
  return returns.groupby(groupby).sum()

def aggregate_returns(returns, period=None, compounded=True):
  """Aggregates returns based on date periods"""
  if period is None or 'day' in period:
    return returns
  index = returns.index

  if 'month' in period:
    return group_returns(returns, index.month, compounded=compounded)

  if 'quarter' in period:
    return group_returns(returns, index.quarter, compounded=compounded)

  if period == "Y" or any(x in period for x in ['year', 'eoy', 'yoy']):
    return group_returns(returns, index.year, compounded=compounded)

  if 'week' in period:
    return group_returns(returns, index.isocalendar().week, compounded=compounded)

  if 'eow' in period or period == "W":
    return group_returns(returns, [index.year, index.isocalendar().week],
                            compounded=compounded)

  if 'eom' in period or period == "M":
    return group_returns(returns, [index.year, index.month],
                            compounded=compounded)

  if 'eoq' in period or period == "Q":
    return group_returns(returns, [index.year, index.quarter],
                            compounded=compounded)

  if not isinstance(period, str):
    return group_returns(returns, period, compounded)

  return returns

def count_consecutive(data):
  """Counts consecutive data"""
  if isinstance(data, pd.DataFrame):
    results = pd.DataFrame()
    for c in data:
      results[c] = data[c] * (data[c].groupby((data[c] != data[c].shift(1)).cumsum()).cumcount() + 1)
    return results

  return data * (data.groupby((data != data.shift(1)).cumsum()).cumcount() + 1)

def to_drawdown_series(returns):
    """Convert returns series to drawdown series"""
    prices = to_quotes(returns, 1)
    prices = clean(prices)
    dd = prices / np.maximum.accumulate(prices) - 1.
    return dd.replace([np.inf, -np.inf, -0], 0)

def remove_outliers(returns, quantile=.95):
  """Returns series of returns without the outliers"""
  return returns[returns < returns.quantile(quantile)]

def exposure(returns):
  """Returns the market exposure time (returns != 0)"""

  return 100 * returns[(returns != 0)].count() / returns.count() / 100

# end


import pandas as pd
from . import utils
from . import stats
from . import plots
from tabulate import tabulate
import io
import re as regex
from base64 import b64encode

def _download_html(html, filename="tearsheet.html"):
    jscode = regex.sub(' +', ' ', """<script>
    var bl=new Blob(['{{html}}'],{type:"text/html"});
    var a=document.createElement("a");
    a.href=URL.createObjectURL(bl);
    a.download="{{filename}}";
    a.hidden=true;document.body.appendChild(a);
    a.innerHTML="download report";
    a.click();</script>""".replace('\n', ''))
    jscode = jscode.replace('{{html}}', _regex.sub(
        ' +', ' ', html.replace('\n', '')))

def _file_stream():
    """Returns a file stream"""
    return io.BytesIO()
  
def _html_table(obj, showindex="default"):
    obj = tabulate(obj.T, headers="keys", tablefmt='html',
                    floatfmt=".2f", showindex=showindex)
    obj = obj.replace(' style="text-align: right;"', '')
    obj = obj.replace(' style="text-align: left;"', '')
    obj = obj.replace(' style="text-align: center;"', '')
    obj = regex.sub('<td> +', '<td>', obj)
    obj = regex.sub(' +</td>', '</td>', obj)
    obj = regex.sub('<th> +', '<th>', obj)
    obj = regex.sub(' +</th>', '</th>', obj)
    return obj

def _embed_figure(figfile, figfmt):
    figbytes = figfile.getvalue()
    if figfmt == 'svg':
        return figbytes.decode()
    data_uri = b64encode(figbytes).decode()
    return '<img src="data:image/{};base64,{}" />'.format(figfmt, data_uri)

def report(returns, benchmark, rf,
         grayscale=False,
         title='Strategy Tearsheet',
         output=None,
         periods=252,
         download_filename='tearsheet.html',
         firm='Tera Capital',
         version='0',
         figfmt='svg',
         template_path=None):

    tpl = ""
    with open(template_path or __file__[:-4] + '.html') as f:
        tpl = f.read()
        f.close()

    date_range = returns.index.strftime('%e %b, %Y')
    tpl = tpl.replace('{{date_range}}', date_range[0] + ' - ' + date_range[-1])
    tpl = tpl.replace('{{title}}', title)
    tpl = tpl.replace('{{firm}}', firm)
    tpl = tpl.replace('{{v}}', version)

    mtrx = report_metrics(returns, benchmark, rf,
                          display=False,
                          periods=periods)

    tpl = tpl.replace('{{info}}', _html_table(mtrx[0]))
    tpl = tpl.replace('{{returns}}', _html_table(mtrx[1]))
    tpl = tpl.replace('{{performance}}', _html_table(mtrx[2]))
    tpl = tpl.replace('{{risk}}', _html_table(mtrx[3]))
    tpl = tpl.replace('{{pnl}}', _html_table(mtrx[4]))

    tpl = tpl.replace('<tr><td></td><td></td><td></td></tr>',
                      '<tr><td colspan="3"><hr></td></tr>')
    tpl = tpl.replace('<tr><td></td><td></td></tr>',
                      '<tr><td colspan="2"><hr></td></tr>')
    dd_info = mtrx[5][mtrx[5].columns.get_level_values(0).unique()[0]]
    dd_info = dd_info.sort_values(
        by='max drawdown', ascending=True)[:10]
    dd_info = dd_info[['start', 'valley', 'end', 'max drawdown', 'days']]
    dd_info.columns = ['Started', 'Valley','Recovered', 'Drawdown', 'Days']
    tpl = tpl.replace('{{dd_info}}', _html_table(dd_info.T, False))

    figs = report_plots(returns.iloc[:,0], benchmark.iloc[:,0], rf, grayscale=grayscale, periods=periods, display=False, save=True)

    tpl = tpl.replace('{{cum_returns}}', figs[0])
    tpl = tpl.replace('{{log_returns}}', figs[1])
    tpl = tpl.replace('{{vol_returns}}', figs[2])
    tpl = tpl.replace('{{eoy_returns}}', figs[3])
    tpl = tpl.replace('{{monthly_dist}}', figs[4])
    tpl = tpl.replace('{{daily_returns}}', figs[5])
    tpl = tpl.replace('{{rolling_beta}}', figs[6])
    tpl = tpl.replace('{{rolling_vol}}', figs[7])
    tpl = tpl.replace('{{rolling_sharpe}}', figs[8])
    tpl = tpl.replace('{{rolling_sortino}}', figs[9])
    tpl = tpl.replace('{{dd_periods}}', figs[10])
    tpl = tpl.replace('{{dd_plot}}', figs[11])
    tpl = tpl.replace('{{monthly_heatmap}}', figs[12])
    tpl = tpl.replace('{{returns_dist}}', figs[13])

    tpl = regex.sub(r'\{\{(.*?)\}\}', '', tpl)
    tpl = tpl.replace('white-space:pre;', '')

    if output is None:
        _download_html(tpl, download_filename)
        return

    with open(output, 'w', encoding='utf-8') as f:
        f.write(tpl)
      
def report_metrics(dr, bench, rf, periods=252, display=True):

  rfd = rf['dR'].rename('DI')
  rfy = rf['yR'].rename('DI')
  frames = [dr, bench]
  df = pd.concat(frames, axis=1)

  dd = utils.to_drawdown_series(df)
  dd_details = utils.drawdown_details(dd)

  dsp_info = pd.DataFrame()
  date_start = dr.index.strftime('%Y-%m-%d')[0]
  date_end = dr.index.strftime('%Y-%m-%d')[-1]
  dsp_info['Strategy'] = pd.Series(dr.columns[0])
  dsp_info['Benchmark'] = pd.Series(bench.columns[0])
  dsp_info['Start Period'] = pd.Series(date_start)
  dsp_info['End Period'] = pd.Series(date_end)
  dsp_info['Working days'] = pd.Series(len(dr.index)) 
  dsp_info['Years'] = pd.Series("{:.1f}".format(len(dr.index)/ periods))

  dsp_returns = pd.DataFrame()
  frames = [dr, bench, rfd ]
  df = pd.concat(frames, axis=1)
  total_return = utils.to_pct(stats.total_return(df),2)
  dsp_returns['Total return'] = pd.Series(total_return)
  cagr = utils.to_pct(stats.cagr(df),2)
  dsp_returns['CAGR'] = pd.Series(cagr)
  d_r = utils.to_pct(stats.expected_return(df),2)
  dsp_returns['E[Daily Returns]'] = pd.Series(d_r)
  m_r = utils.to_pct(stats.expected_return(df, aggregate='M'),2)
  dsp_returns['E[Monthly Returns]'] = pd.Series(m_r)
  y_r = utils.to_pct(stats.expected_return(df, aggregate='Y'),2)
  dsp_returns['E[Yearly Returns]'] = pd.Series(y_r)
  r_mtd = utils.to_pct(stats.total_return(utils.mtd(df, df.index[-1])),2)
  dsp_returns['MTD'] = pd.Series(r_mtd)
  r_l3m = utils.to_pct(stats.total_return(utils.l3m(df, df.index[-1])),2)
  dsp_returns['3M'] = pd.Series(r_l3m)
  r_l6m = utils.to_pct(stats.total_return(utils.l6m(df, df.index[-1])),2)
  dsp_returns['6M'] = pd.Series(r_l6m)
  r_ytd = utils.to_pct(stats.total_return(utils.ytd(df, df.index[-1])),2)
  dsp_returns['YTD'] = pd.Series(r_ytd)
  r_l1y = utils.to_pct(stats.total_return(utils.l1y(df, df.index[-1])),2)
  dsp_returns['1Y'] = pd.Series(r_l1y)
  cagr_3y = utils.to_pct(stats.cagr(utils.l3y(df, df.index[-3])),2)
  dsp_returns['3Y (aa)'] = pd.Series(cagr_3y)
  cagr_5y = utils.to_pct(stats.cagr(utils.l5y(df, df.index[-1])),2)
  dsp_returns['5Y (aa)'] = pd.Series(cagr_5y)
  cagr_10y = utils.to_pct(stats.cagr(utils.l10y(df, df.index[-1])),2)
  dsp_returns['10Y (aa)'] = pd.Series(cagr_10y)
  dsp_returns['Inception (aa)'] = pd.Series(cagr)
  
  dsp_performance = pd.DataFrame()
  frames = [dr, bench]
  df = pd.concat(frames, axis=1)
  sharpe = utils.to_num(stats.sharpe(df, rfy, smart=False),2)
  dsp_performance['Sharpe'] = pd.Series(sharpe)
  smart_sharpe = utils.to_num(stats.sharpe(df, rfy, smart=True),2)
  dsp_performance['Smart Sharpe'] = pd.Series(smart_sharpe)
  sortino = utils.to_num(stats.sortino(df, rfy, smart=False),2)
  dsp_performance['Sortino'] = pd.Series(sortino)
  smart_sortino = utils.to_num(stats.sortino(df, rfy, smart=True),2)
  dsp_performance['Smart Sortino'] = pd.Series(smart_sortino)
  adj_sortino = utils.to_num(stats.adjusted_sortino(df, rfy, smart=False),2)
  dsp_performance['Sortino/√2'] = pd.Series(adj_sortino)
  adj_smart_sortino = utils.to_num(stats.adjusted_sortino(df, rfy, smart=True),2)
  dsp_performance['Smart Sortino/√2'] = pd.Series(adj_smart_sortino)
  omega = utils.to_num(stats.omega(df, rfy),2)
  dsp_performance['Omega'] = pd.Series(omega)
  beta = utils.to_num(stats.beta(dr, bench),2)
  dsp_performance['Beta'] = pd.Series(beta)
  alpha = utils.to_num(stats.alpha(dr, bench),2)
  dsp_performance['Alpha'] = pd.Series(alpha)
  
  dsp_risk = pd.DataFrame()
  frames = [dr, bench]
  df = pd.concat(frames, axis=1)
  max_drawdown = utils.to_pct(stats.max_drawdown(df),2)
  dsp_risk['Max Drawdown'] = pd.Series(max_drawdown)
  longest_drawdown_days = utils.to_num(stats.longest_drawdown_days(df),0)
  dsp_risk['Longest DD Days'] = pd.Series(longest_drawdown_days)
  avg_drawdown = utils.to_pct(stats.avg_drawdown(dd_details),2)
  dsp_risk['Avg. Drawdown'] = pd.Series(avg_drawdown)
  avg_drawdown_days = utils.to_num(stats.avg_drawdown_days(df),0)
  dsp_risk['Avg. Drawdown Days'] = pd.Series(avg_drawdown_days)
  volatility = utils.to_pct(stats.volatility(df, periods),2)
  dsp_risk['Volatility (aa)'] = pd.Series(volatility)
  r_squared = utils.to_num(stats.r_squared(dr, bench),2)
  dsp_risk['R^2'] = pd.Series(r_squared)
  calmar = utils.to_num(stats.calmar(df),2)
  dsp_risk['Calmar'] = pd.Series(calmar)
  skew = utils.to_num(stats.skew(df),2)
  dsp_risk['Skew'] = pd.Series(skew)
  kurtosis = utils.to_num(stats.kurtosis(df),2)
  dsp_risk['Kurtosis'] = pd.Series(kurtosis)
  kelly_criterion = utils.to_pct(stats.kelly_criterion(df),2)
  dsp_risk['Kelly Criterion'] = pd.Series(kelly_criterion)
  risk_of_ruin = utils.to_pct(stats.risk_of_ruin(df),2)
  dsp_risk['Risk of Ruin'] = pd.Series(risk_of_ruin)
  value_at_risk = utils.to_pct(stats.value_at_risk(df),2)
  dsp_risk['Daily VaR'] = pd.Series(value_at_risk)
  conditional_value_at_risk = utils.to_pct(stats.conditional_value_at_risk(df),2)
  dsp_risk['Expected Shortfall (cVaR)'] = pd.Series(conditional_value_at_risk)
  recovery_factor = utils.to_num(stats.recovery_factor(df),2)
  dsp_risk['Recovery Factor'] = pd.Series(recovery_factor)
  ulcer_index = utils.to_num(stats.ulcer_index(df),2)
  dsp_risk['Ulcer Index'] = pd.Series(ulcer_index)
  serenity_index = utils.to_num(stats.serenity_index(df, rfy),2)
  dsp_risk['Serenity Index'] = pd.Series(serenity_index)
  
  dsp_pnl = pd.DataFrame()
  frames = [dr, bench]
  df = pd.concat(frames, axis=1)
  er = utils.to_excess_returns(df, rfy)
  gtp = utils.to_num(stats.gain_to_pain_ratio(er),2)
  dsp_pnl['Gain/Pain Ratio'] = pd.Series(gtp)
  gtp_1m = utils.to_num(stats.gain_to_pain_ratio(er, 'M'),2)
  dsp_pnl['Gain/Pain (1M)'] = pd.Series(gtp_1m)
  gtp_3m = utils.to_num(stats.gain_to_pain_ratio(er, 'Q'),2)
  dsp_pnl['Gain/Pain (3M)'] = pd.Series(gtp_3m)
  gtp_6m = utils.to_num(stats.gain_to_pain_ratio(er, 'BQ'),2)
  dsp_pnl['Gain/Pain (6M)'] = pd.Series(gtp_6m)
  gtp_1y = utils.to_num(stats.gain_to_pain_ratio(er, 'A'),2)
  dsp_pnl['Gain/Pain (1Y)'] = pd.Series(gtp_1y)
  payoff_ratio = utils.to_num(stats.payoff_ratio(df),2)
  dsp_pnl['Payoff Ratio'] = pd.Series(payoff_ratio)
  profit_factor = utils.to_num(stats.profit_factor(df),2)
  dsp_pnl['Profit Factor'] = pd.Series(profit_factor)
  common_sense_ratio = utils.to_num(stats.common_sense_ratio(df),2)
  dsp_pnl['Common Sense Ratio'] = pd.Series(common_sense_ratio)
  cpc_index = utils.to_num(stats.cpc_index(df),2)
  dsp_pnl['CPC Index'] = pd.Series(cpc_index)
  tail_ratio = utils.to_num(stats.tail_ratio(df),2)
  dsp_pnl['Tail Ratio'] = pd.Series(tail_ratio)
  outlier_win_ratio = utils.to_num(stats.outlier_win_ratio(df),2)
  dsp_pnl['Outlier Win Ratio'] = pd.Series(outlier_win_ratio)
  outlier_loss_ratio = utils.to_num(stats.outlier_loss_ratio(df),2)
  dsp_pnl['Outlier Loss Ratio'] = pd.Series(outlier_loss_ratio)
  avg_win = utils.to_pct(stats.avg_win(df, aggregate='M'),2)
  dsp_pnl['Avg. Up Month'] = pd.Series(avg_win)
  avg_loss = utils.to_pct(stats.avg_loss(df, aggregate='M'),2)
  dsp_pnl['Avg. Down Month'] = pd.Series(avg_loss)
  win_rate = utils.to_pct(stats.win_rate(df),2)
  dsp_pnl['Win Days %'] = pd.Series(win_rate)
  win_rate_m = utils.to_pct(stats.win_rate(df, aggregate='M'),2)
  dsp_pnl['Win Month %'] = pd.Series(win_rate_m)
  win_rate_q = utils.to_pct(stats.win_rate(df, aggregate='Q'),2)
  dsp_pnl['Win Quarter %'] = pd.Series(win_rate_q)
  win_rate_y = utils.to_pct(stats.win_rate(df, aggregate='Y'),2)
  dsp_pnl['Win Year %'] = pd.Series(win_rate_y)
  best = utils.to_pct(stats.best(df),2)
  dsp_pnl['Best Day'] = pd.Series(best)
  worst = utils.to_pct(stats.worst(df),2)
  dsp_pnl['Worst Day'] = pd.Series(worst)
  best_m = utils.to_pct(stats.best(df, aggregate='M'),2)
  dsp_pnl['Best Month'] = pd.Series(best_m)
  worst_m = utils.to_pct(stats.worst(df, aggregate='M'),2)
  dsp_pnl['Worst Month'] = pd.Series(worst_m)
  best_y = utils.to_pct(stats.best(df, aggregate='Y'),2)
  dsp_pnl['Best Year'] = pd.Series(best_y)
  worst_y = utils.to_pct(stats.worst(df, aggregate='Y'),2)
  dsp_pnl['Worst Year'] = pd.Series(worst_y)

  if display:
    print('[Analysis info]')
    print(tabulate(dsp_info.T, tablefmt='simple'))
    print('\n')
    print('[Performance Metrics]')
    print('Return Info')
    print(tabulate(dsp_returns.T, headers="keys", tablefmt='simple'))
    print('\n')
    print('Performance Info')
    print(tabulate(dsp_performance.T, headers="keys", tablefmt='simple'))
    print('\n')
    print('Risk Info')
    print(tabulate(dsp_risk.T, headers="keys", tablefmt='simple'))
    print('\n')
    print('P&L Info')
    print(tabulate(dsp_pnl.T, headers="keys", tablefmt='simple'))
    print('\n')
    for c in dd_details.columns.get_level_values(0).unique():
      print('%s : 5 Worst Drawdowns' % c)
      print(tabulate(dd_details[c].sort_values(by='max drawdown', ascending=True)[:5], headers="keys", tablefmt='simple', showindex=False))
      print('\n')
    return None

  return dsp_info, dsp_returns, dsp_performance, dsp_risk, dsp_pnl, dd_details

def report_plots(returns, benchmark, rf, grayscale=False, figsize=(8, 5), periods=252, display=True, figfmt='svg', save=False):
  rfd=rf['dR']
  rfy=rf['yR']
  
  figs = []
  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_returns(returns, benchmark, rfd,
              grayscale=grayscale,
              figsize=(figsize[0], figsize[0]*.6),
              show=display,
              savefig=savefig,
              ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_log_returns(returns, benchmark, rfd,
              grayscale=grayscale,
              figsize=(figsize[0], figsize[0]*.5),
              show=display,
              savefig=savefig,
              ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_volmatch_returns(returns, benchmark, rfd,
              grayscale=grayscale,
              figsize=(figsize[0], figsize[0]*.5),
              show=display,
              savefig=savefig,
              ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None
  
  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_yearly_returns(returns, benchmark, rfd,
              grayscale=grayscale,
              figsize=(figsize[0], figsize[0]*.5),
              show=display,
              savefig=savefig,
              ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_histogram(returns,
                grayscale=grayscale,
                figsize=(figsize[0], figsize[0]*.5),
                show=display,
                savefig=savefig,
                ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_daily_returns(returns,
                    grayscale=grayscale,
                    figsize=(figsize[0], figsize[0]*.3),
                    show=display,
                    savefig=savefig,
                    ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_rolling_beta(returns, benchmark,
                   grayscale=grayscale,
                   window1=int(periods/2),
                   window2=periods,
                   figsize=(figsize[0], figsize[0]*.3),
                   show=display,
                   savefig=savefig,
                   ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None
  
  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_rolling_volatility(returns, benchmark,
                         grayscale=grayscale,
                         figsize=(figsize[0], figsize[0]*.3),
                         show=display,
                         savefig=savefig,
                         ylabel=False,
                         period=int(periods/2))
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_rolling_sharpe(returns, benchmark, rfy,
                     grayscale=grayscale,
                     figsize=(figsize[0], figsize[0]*.3),
                     show=display,
                     savefig=savefig,
                     ylabel=False,
                     period=int(periods/2))
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None
  
  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_rolling_sortino(returns, benchmark, rfy,
                      grayscale=grayscale,
                      figsize=(figsize[0], figsize[0]*.3),
                      show=display,
                      savefig=savefig,
                      ylabel=False,
                      period=int(periods/2))
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_drawdowns_periods(returns, periods=5,
                        grayscale=grayscale,
                        figsize=(figsize[0], figsize[0]*.5),
                        show=display,
                        savefig=savefig,
                        ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_drawdown(returns,
               grayscale=grayscale,
               figsize=(figsize[0], figsize[0]*.4),
               show=display,
               savefig=savefig,
               ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_monthly_heatmap(returns,
                      grayscale=grayscale,
                      figsize=(figsize[0], figsize[0]*.5),
                      show=display,
                      savefig=savefig,
                      ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  savefig = {'fname': _file_stream(), 'format': figfmt} if save else None
  plots.plt_distribution(returns,
                   grayscale=grayscale,
                   figsize=(figsize[0], figsize[0]*.5),
                   show=display,
                   savefig=savefig,
                   ylabel=False)
  figs.append(_embed_figure(savefig['fname'], savefig['format'])) if save else None

  if save:
    return figs



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# FinTools: Portfolio analytics
# https://github.com/zegoverno/zenfin
#
# Copyright 2021 José Governo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import version

__version__ = version.version
__author__ = "José Governo"


from . import stats, utils, reports

__all__ = ['stats', 'reports', 'utils']


import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.ticker import FormatStrFormatter, FuncFormatter

import pandas as pd
import numpy as np

import seaborn as sns

sns.set(font_scale=1.1, rc={
    'figure.figsize': (10, 6),
    'axes.facecolor': 'white',
    'figure.facecolor': 'white',
    'grid.color': '#dddddd',
    'grid.linewidth': 0.5,
    "lines.linewidth": 1.5,
    'text.color': '#333333',
    'xtick.color': '#666666',
    'ytick.color': '#666666'
})

_FLATUI_COLORS = [ "#348dc1", "#4fa487", "#af4b64",
                  "#fedd78", "#9b59b6", "#808080"]
_GRAYSCALE_COLORS = ['silver', '#222222', 'gray'] * 3

def _get_colors(grayscale):
    colors = _FLATUI_COLORS
    ls = '-'
    alpha = .8
    if grayscale:
        colors = _GRAYSCALE_COLORS
        ls = '-'
        alpha = 0.5
    return colors, ls, alpha

def format_pct_axis(x, _):
    x *= 100  # lambda x, loc: "{:,}%".format(int(x * 100))
    if x >= 1e12:
        res = '%1.1fT%%' % (x * 1e-12)
        return res.replace('.0T%', 'T%')
    if x >= 1e9:
        res = '%1.1fB%%' % (x * 1e-9)
        return res.replace('.0B%', 'B%')
    if x >= 1e6:
        res = '%1.1fM%%' % (x * 1e-6)
        return res.replace('.0M%', 'M%')
    if x >= 1e3:
        res = '%1.1fK%%' % (x * 1e-3)
        return res.replace('.0K%', 'K%')
    res = '%1.0f%%' % x
    return res.replace('.0%', '%')
  
def timeseries(returns, benchmark=None, rf=0.,
                    title="Returns",
                    fill=False,
                    returns_label="Strategy",
                    hline=None,
                    hlw=None,
                    hlcolor="red",
                    hllabel="",
                    percent=True,
                    log_scale=False,
                    lw=1.5,
                    figsize=(10, 6),
                    ylabel="",
                    grayscale=False,
                    fontname="Arial",
                    subtitle=True,
                    savefig=None,
                    show=True):

    colors, ls, alpha = _get_colors(grayscale)

    fig, ax = plt.subplots(figsize=figsize)

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    fig.suptitle(title+"\n", y=.99, fontweight="bold", fontname=fontname,
                 fontsize=14, color="black")

    if subtitle:
        ax.set_title("\n%s - %s                  " % (
            returns.index.date[:1][0].strftime('%e %b \'%y'),
            returns.index.date[-1:][0].strftime('%e %b \'%y')
        ), fontsize=12, color='gray')

    fig.set_facecolor('white')
    ax.set_facecolor('white')

    if isinstance(rf, pd.Series):
      ax.plot(rf, lw=lw, ls=ls, label="Risk-Free", color=colors[2])

    if isinstance(benchmark, pd.Series):
      ax.plot(benchmark, lw=lw, ls=ls, label="Benchmark", color=colors[1])

    alpha = .25 if grayscale else 1
    ax.plot(returns, lw=lw, label=returns_label, color=colors[0], alpha=alpha)

    if fill:
        ax.fill_between(returns.index, 0, returns, color=colors[0], alpha=.25)

    fig.autofmt_xdate()

    ax.fmt_xdata = mdates.DateFormatter('%Y-%m-%d')

    if hline:
        ax.legend(fontsize=12)
        if grayscale:
            hlcolor = 'black'
        ax.axhline(hline, ls="--", lw=hlw, color=hlcolor,
                   label=hllabel, zorder=2)

    ax.axhline(0, ls="-", lw=1,
               color='gray', zorder=1)
    ax.axhline(0, ls="--", lw=1,
               color='white' if grayscale else 'black', zorder=2)

    plt.yscale("symlog" if log_scale else "linear")

    if percent:
        ax.yaxis.set_major_formatter(FuncFormatter(format_pct_axis))

    ax.set_xlabel('')
    if ylabel:
        ax.set_ylabel(ylabel, fontname=fontname,
                      fontweight='bold', fontsize=12, color="black")
    ax.yaxis.set_label_coords(-.1, .5)

    ax.legend(fontsize=12)
    
    try:
        plt.subplots_adjust(hspace=0, bottom=0, top=1)
    except Exception:
        pass

    try:
        fig.tight_layout()
    except Exception:
        pass

    if savefig:
        if isinstance(savefig, dict):
            plt.savefig(**savefig)
        else:
            plt.savefig(savefig)

    if show:
        plt.show(block=False)

    plt.close()

    if not show:
        return fig

    return None
                      
def returns_bars(returns, benchmark, rf,
                      title="Returns",
                      fill=False,
                      returns_label="Strategy",
                      hline=None, hlw=None, hlcolor="red", hllabel="",
                      log_scale=False,
                      lw=1.5,
                      figsize=(10, 6),
                      ylabel="",
                      grayscale=False,
                      fontname="Arial",
                      subtitle=True,
                      savefig=None,
                      show=True):

    df = pd.DataFrame(index=returns.index, data={returns_label: returns,
                                                 'Benchmark': benchmark,
                                                 'Risk-free': rf
                                                 })
    colors, ls, alpha = _get_colors(grayscale)

    fig, ax = plt.subplots(figsize=figsize)

    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    fig.suptitle(title+"\n", y=.99, fontweight="bold", fontname=fontname,
                 fontsize=14, color="black")

    if subtitle:
        ax.set_title("\n%s - %s                  " % (
            returns.index.date[:1][0].strftime('%e %b \'%y'),
            returns.index.date[-1:][0].strftime('%e %b \'%y')
        ), fontsize=12, color='gray')

    df.plot(kind='bar', ax=ax, color=colors)

    fig.set_facecolor('white')
    ax.set_facecolor('white')

    try:
        ax.set_xticklabels(df.index.year)
        years = sorted(list(set(df.index.year)))
    except AttributeError:
        ax.set_xticklabels(df.index)
        years = sorted(list(set(df.index)))

    # rotate and align the tick labels so they look better
    fig.autofmt_xdate()

    if hline:
        if grayscale:
            hlcolor = 'gray'
        ax.axhline(hline, ls="--", lw=hlw, color=hlcolor,
                   label=hllabel, zorder=2)

    ax.axhline(0, ls="--", lw=1, color="#000000", zorder=2)

    ax.legend(fontsize=12)

    plt.yscale("symlog" if log_scale else "linear")

    ax.set_xlabel('')
    if ylabel:
        ax.set_ylabel("Returns", fontname=fontname,
                      fontweight='bold', fontsize=12, color="black")
        ax.yaxis.set_label_coords(-.1, .5)

    ax.yaxis.set_major_formatter(FuncFormatter(format_pct_axis))

    try:
        plt.subplots_adjust(hspace=0, bottom=0, top=1)
    except Exception:
        pass

    try:
        fig.tight_layout()
    except Exception:
        pass

    if savefig:
        if isinstance(savefig, dict):
            plt.savefig(**savefig)
        else:
            plt.savefig(savefig)

    if show:
        plt.show(block=False)

    plt.close()

    if not show:
        return fig

    return None

def histogram(returns,
                   bins=20,
                   fontname='Arial', 
                   grayscale=False,
                   title="Returns", 
                   kde=True, 
                   figsize=(10, 6),
                   ylabel=True,
                   subtitle=True,
                   savefig=None,
                   show=True):

    colors = ['#348dc1', '#003366', 'red']
    if grayscale:
        colors = ['silver', 'gray', 'black']

    fig, ax = plt.subplots(figsize=figsize)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    fig.suptitle(title+"\n", y=.99, fontweight="bold", fontname=fontname,
                 fontsize=14, color="black")

    if subtitle:
        ax.set_title("\n%s - %s                   " % (
            returns.index.date[:1][0].strftime('%Y'),
            returns.index.date[-1:][0].strftime('%Y')
        ), fontsize=12, color='gray')

    fig.set_facecolor('white')
    ax.set_facecolor('white')

    ax.axvline(returns.mean(), ls="--", lw=1.5,
               color=colors[2], zorder=2, label="Average")

    sns.histplot(returns, bins=bins,
                  color=colors[0],
                  alpha=1,
                  kde=kde,
                  stat="density",
                  ax=ax)
    
    sns.kdeplot(returns, color='black', linewidth=1.5)

    ax.xaxis.set_major_formatter(plt.FuncFormatter(
        lambda x, loc: "{:,}%".format(int(x*100))))

    ax.axhline(0.01, lw=1, color="#000000", zorder=2)
    ax.axvline(0, lw=1, color="#000000", zorder=2)

    ax.set_xlabel('')
    if ylabel:
        ax.set_ylabel("Occurrences", fontname=fontname,
                      fontweight='bold', fontsize=12, color="black")
        ax.yaxis.set_label_coords(-.1, .5)

    ax.legend(fontsize=12)

    try:
        plt.subplots_adjust(hspace=0, bottom=0, top=1)
    except Exception:
        pass

    try:
        fig.tight_layout()
    except Exception:
        pass

    if savefig:
        if isinstance(savefig, dict):
            plt.savefig(**savefig)
        else:
            plt.savefig(savefig)

    if show:
        plt.show(block=False)

    plt.close()

    if not show:
        return fig

    return None

def rolling_beta(beta1, beta2,
                      window1_label="",
                      window2_label="",
                      title="", 
                      hlcolor="red",
                      figsize=(10, 6),
                      grayscale=False,
                      fontname="Arial",
                      lw=1.5,
                      ylabel=True,
                      subtitle=True,
                      savefig=None,
                      show=True):

    colors, _, _ = _get_colors(grayscale)

    fig, ax = plt.subplots(figsize=figsize)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    fig.suptitle(title+"\n", y=.99, fontweight="bold", fontname=fontname,
                 fontsize=14, color="black")

    if subtitle:
        ax.set_title("\n%s - %s                   " % (
            beta1.index.date[:1][0].strftime('%e %b \'%y'),
            beta1.index.date[-1:][0].strftime('%e %b \'%y')
        ), fontsize=12, color='gray')

    ax.plot(beta1, lw=lw, label=window1_label, color=colors[1])
    ax.plot(beta2,lw=lw, label=window2_label, color="gray", alpha=0.8)

    mmin = min([-100, int(beta1.min()*100)])
    mmax = max([100, int(beta1.max()*100)])
    step = 50 if (mmax-mmin) >= 200 else 100
    ax.set_yticks([x / 100 for x in list(range(mmin, mmax, step))])

    hlcolor = 'black' if grayscale else hlcolor
    ax.axhline(beta1.mean(), ls="--", lw=1.5,
               color=hlcolor, zorder=2)

    ax.axhline(0, ls="--", lw=1, color="#000000", zorder=2)

    fig.autofmt_xdate()

    ax.fmt_xdata = mdates.DateFormatter('%Y-%m-%d')

    if ylabel:
        ax.set_ylabel("Beta", fontname=fontname,
                      fontweight='bold', fontsize=12, color="black")
        ax.yaxis.set_label_coords(-.1, .5)

    ax.legend(fontsize=12)
    try:
        plt.subplots_adjust(hspace=0, bottom=0, top=1)
    except Exception:
        pass

    try:
        fig.tight_layout()
    except Exception:
        pass

    if savefig:
        if isinstance(savefig, dict):
            plt.savefig(**savefig)
        else:
            plt.savefig(savefig)

    if show:
        plt.show(block=False)

    plt.close()

    if not show:
        return fig

    return None

def rolling_stats(returns, benchmark=None,
                       title="",
                       returns_label="Strategy",
                       hline=None,
                       hlw=None,
                       hlcolor="red",
                       hllabel="",
                       lw=1.5,
                       figsize=(10, 6),
                       ylabel="",
                       grayscale=False,
                       fontname="Arial",
                       subtitle=True,
                       savefig=None,
                       show=True):

    colors, _, _ = _get_colors(grayscale)

    fig, ax = plt.subplots(figsize=figsize)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    
    if isinstance(benchmark, pd.Series):
        ax.plot(benchmark, lw=lw, label="Benchmark", color=colors[1], alpha=.8)

    ax.plot(returns, lw=lw, label=returns_label, color=colors[0])

    fig.autofmt_xdate()

    fig.suptitle(title+"\n", y=.99, fontweight="bold", fontname=fontname,
                 fontsize=14, color="black")

    if subtitle:
        ax.set_title("\n%s - %s                   " % (
            returns.index.date[:1][0].strftime('%e %b \'%y'),
            returns.index.date[-1:][0].strftime('%e %b \'%y')
        ), fontsize=12, color='gray')

    if hline:
        if grayscale:
            hlcolor = 'black'
        ax.axhline(hline, ls="--", lw=hlw, color=hlcolor,
                   label=hllabel, zorder=2)

    ax.axhline(0, ls="--", lw=1, color="#000000", zorder=2)

    if ylabel:
        ax.set_ylabel(ylabel, fontname=fontname,
                      fontweight='bold', fontsize=12, color="black")
        ax.yaxis.set_label_coords(-.1, .5)

    ax.yaxis.set_major_formatter(FormatStrFormatter('%.2f'))

    ax.legend(fontsize=12)

    try:
        plt.subplots_adjust(hspace=0, bottom=0, top=1)
    except Exception:
        pass

    try:
        fig.tight_layout()
    except Exception:
        pass

    if savefig:
        if isinstance(savefig, dict):
            plt.savefig(**savefig)
        else:
            plt.savefig(savefig)
    if show:
        plt.show(block=False)
    plt.close()


    if not show:
        return fig

    return None

def longest_drawdowns(returns, longest_dd, periods,
                           lw=1.5,
                           fontname='Arial',
                           grayscale=False,
                           log_scale=False,
                           figsize=(10, 6),
                           ylabel=True,
                           subtitle=True,
                           savefig=None, 
                           show=True):

    colors = ['#348dc1', '#003366', 'red']
    if grayscale:
        colors = ['#000000'] * 3

    fig, ax = plt.subplots(figsize=figsize)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    fig.suptitle("Worst %.0f Drawdown Periods\n" %
                 periods, y=.99, fontweight="bold", fontname=fontname,
                 fontsize=14, color="black")
    if subtitle:
        ax.set_title("\n%s - %s                   " % (
            returns.index.date[:1][0].strftime('%e %b \'%y'),
            returns.index.date[-1:][0].strftime('%e %b \'%y')
        ), fontsize=12, color='gray')

    fig.set_facecolor('white')
    ax.set_facecolor('white')
    ax.plot(returns, lw=lw, label="Backtest", color=colors[0])

    highlight = 'black' if grayscale else 'red'
    for _, row in longest_dd.iterrows():
        ax.axvspan(*mdates.datestr2num([str(row['start']), str(row['end'])]),
                   color=highlight, alpha=.1)

  
    fig.autofmt_xdate()

    ax.fmt_xdata = mdates.DateFormatter('%Y-%m-%d')

    ax.axhline(0, ls="--", lw=1, color="#000000", zorder=2)
    plt.yscale("symlog" if log_scale else "linear")
    if ylabel:
        ax.set_ylabel("Cumulative Returns", fontname=fontname,
                      fontweight='bold', fontsize=12, color="black")
        ax.yaxis.set_label_coords(-.1, .5)

    ax.yaxis.set_major_formatter(FuncFormatter(format_pct_axis))

    fig.autofmt_xdate()

    try:
        plt.subplots_adjust(hspace=0, bottom=0, top=1)
    except Exception:
        pass

    try:
        fig.tight_layout()
    except Exception:
        pass

    if savefig:
        if isinstance(savefig, dict):
            plt.savefig(**savefig)
        else:
            plt.savefig(savefig)

    if show:
        plt.show(block=False)

    plt.close()

    if not show:
        return fig

    return None

def distribution(port,
                      figsize=(10, 6),
                      fontname='Arial',
                      grayscale=False,
                      ylabel=True,
                      subtitle=True,
                      compounded=True,
                      savefig=None,
                      show=True):

    colors = _FLATUI_COLORS
    if grayscale:
        colors = ['#f9f9f9', '#dddddd', '#bbbbbb', '#999999', '#808080']
    # colors, ls, alpha = _get_colors(grayscale)

    
    fig, ax = plt.subplots(figsize=figsize)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    fig.suptitle("Return Quantiles\n", y=.99,
                 fontweight="bold", fontname=fontname,
                 fontsize=14, color="black")

    if subtitle:
        ax.set_title("\n%s - %s                   " % (
            port.index.date[:1][0].strftime('%e %b \'%y'),
            port.index.date[-1:][0].strftime('%e %b \'%y')
        ), fontsize=12, color='gray')

    fig.set_facecolor('white')
    ax.set_facecolor('white')

    sns.boxplot(data=port, ax=ax, palette=tuple(colors[:5]))

    ax.yaxis.set_major_formatter(plt.FuncFormatter(
        lambda x, loc: "{:,}%".format(int(x*100))))

    if ylabel:
        ax.set_ylabel('Rerurns', fontname=fontname,
                      fontweight='bold', fontsize=12, color="black")
        ax.yaxis.set_label_coords(-.1, .5)

    fig.autofmt_xdate()

    try:
        plt.subplots_adjust(hspace=0)
    except Exception:
        pass
    try:
        fig.tight_layout(w_pad=0, h_pad=0)
    except Exception:
        pass

    if savefig:
        if isinstance(savefig, dict):
            plt.savefig(**savefig)
        else:
            plt.savefig(savefig)

    if show:
        plt.show(block=False)

    plt.close()

    if not show:
        return fig

    return None

def heatmap(returns,
                 title='',
                 annot_size=10,
                 figsize=(10, 5),
                 cbar=True,
                 square=False,
                 eoy=False,
                 grayscale=False,
                 fontname='Arial',
                 ylabel=True,
                 savefig=None,
                 show=True):

    cmap = 'gray' if grayscale else 'RdYlGn'
    fig_height = len(returns) / 3

    if figsize is None:
        size = list(plt.gcf().get_size_inches())
        figsize = (size[0], size[1])

    figsize = (figsize[0], max([fig_height, figsize[1]]))

    if cbar:
        figsize = (figsize[0]*1.04, max([fig_height, figsize[1]]))

    fig, ax = plt.subplots(figsize=figsize)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    fig.set_facecolor('white')
    ax.set_facecolor('white')

    ax.set_title(title, fontsize=14, y=.995,
                 fontname=fontname, fontweight='bold', color='black')

    ax = sns.heatmap(returns, ax=ax, annot=True, center=0,
                      annot_kws={"size": annot_size},
                      fmt="0.2f", linewidths=0.5,
                      square=square, cbar=cbar, cmap=cmap,
                      cbar_kws={'format': '%.0f%%'})

    if ylabel:
        ax.set_ylabel('Years', fontname=fontname,
                      fontweight='bold', fontsize=12)
        ax.yaxis.set_label_coords(-.1, .5)

    ax.tick_params(colors="#808080")
    plt.xticks(rotation=0, fontsize=annot_size*1.2)
    plt.yticks(rotation=0, fontsize=annot_size*1.2)

    try:
        plt.subplots_adjust(hspace=0, bottom=0, top=1)
    except Exception:
        pass
    try:
        fig.tight_layout(w_pad=0, h_pad=0)
    except Exception:
        pass

    if savefig:
        if isinstance(savefig, dict):
            plt.savefig(**savefig)
        else:
            plt.savefig(savefig)

    if show:
        plt.show(block=False)

    plt.close()

    if not show:
        return fig

    return None


import warnings
import matplotlib.pyplot as _plt
from matplotlib.ticker import (
    StrMethodFormatter as StrMethodFormatter,
    FuncFormatter as FuncFormatter
)

import numpy as np
import pandas as pd
import seaborn as sns

from . import stats, _plot, utils, reports

def plt_returns(returns, benchmark, rf, 
                title='Cumulative returns',
                grayscale=False,
                figsize=(10, 6),
                fontname='Arial',
                lw=1.5,
                ylabel=False,
                subtitle=True,
                savefig=None,
                show=True):
  returns = utils.to_rolling_returns(returns)
  benchmark = utils.to_rolling_returns(benchmark)
  rf = utils.to_rolling_returns(rf)
  fig = _plot.timeseries(returns, benchmark, rf, title, ylabel=ylabel, lw=lw, figsize=figsize, fontname=fontname, grayscale=grayscale, subtitle=subtitle, savefig=savefig, show=show )
  if not show:
    return fig 
    
def plt_log_returns(returns, benchmark, rf, 
                title='Cumulative returns (Log scaled)',
                grayscale=False,
                figsize=(10, 6),
                fontname='Arial',
                lw=1.5,
                ylabel=False,
                subtitle=True,
                savefig=None,
                show=True):
  returns = utils.to_rolling_returns(returns)
  benchmark = utils.to_rolling_returns(benchmark)
  rf = utils.to_rolling_returns(rf)
  fig = _plot.timeseries(returns, benchmark, rf, title, log_scale=True, ylabel=ylabel, lw=lw, figsize=figsize, fontname=fontname, grayscale=grayscale, subtitle=subtitle, savefig=savefig, show=show )
  if not show:
    return fig

def plt_volmatch_returns(returns, benchmark, rf, 
                title='Cumulative returns (Volatility Matched)',
                grayscale=False,
                figsize=(10, 6),
                fontname='Arial',
                lw=1.5,
                ylabel=False,
                subtitle=True,
                savefig=None,
                show=True):
  returns = utils.match_volatility(returns, benchmark)
  returns = utils.to_rolling_returns(returns)
  benchmark = utils.to_rolling_returns(benchmark)
  rf = utils.to_rolling_returns(rf)
  fig = _plot.timeseries(returns, benchmark, rf, title, ylabel=ylabel, lw=lw, figsize=figsize, fontname=fontname, grayscale=grayscale, subtitle=subtitle, savefig=savefig, show=show )
  if not show:
    return fig 

def plt_yearly_returns(returns, benchmark, rf, 
                title='EOY Returns',
                grayscale=False,
                figsize=(10, 5),
                fontname='Arial',
                hlw=1.5,
                hlcolor='red',
                hllabel='',
                ylabel=False,
                subtitle=True,
                savefig=None,
                show=True):
  returns = returns.resample('A').apply(stats.total_return)
  benchmark = benchmark.resample('A').apply(stats.total_return)
  rf = rf.resample('A').apply(stats.total_return)
  fig = _plot.returns_bars(returns, benchmark, rf, title, ylabel=ylabel,hline=returns.mean(), hlw=hlw, hllabel=hllabel, hlcolor=hlcolor, figsize=figsize, fontname=fontname, grayscale=grayscale, subtitle=subtitle, savefig=savefig, show=show )
  if not show:
    return fig 

def plt_histogram(returns,
                  resample='M',
                  title='',
                  grayscale=False,
                  figsize=(10, 5),
                  fontname='Arial',
                  ylabel=False,
                  subtitle=True,
                  savefig=None,
                  show=True):
  if resample == 'W':
      title = "Weekly"
  elif resample == 'M':
      title = "Monthly"
  elif resample == 'Q':
      title = "Quarterly"
  elif resample == 'A':
      title = "Annual"
  else:
      title = ""
  returns = returns.resample(resample).apply(stats.total_return)
  fig = _plot.histogram(returns, title="Distribution of %sReturns" % title, ylabel=ylabel, figsize=figsize, fontname=fontname, grayscale=grayscale, subtitle=subtitle, savefig=savefig, show=show )
  if not show:
    return fig

def plt_daily_returns(returns,
                  title="Daily Returns",
                  grayscale=False,
                  figsize=(10, 4),
                  fontname='Arial',
                  lw=0.5,
                  log_scale=False,
                  ylabel=False,
                  subtitle=True,
                  savefig=None,
                  show=True):

    fig = _plot.timeseries(returns, None, 
                          title,
                          ylabel=ylabel,
                          log_scale=log_scale,
                          lw=lw,
                          figsize=figsize,
                          fontname=fontname,
                          grayscale=grayscale,
                          subtitle=subtitle,
                          savefig=savefig,
                          show=show)
    if not show:
        return fig

def plt_rolling_beta(returns, benchmark,
                 window1=126, window1_label="6-Months",
                 window2=252, window2_label="12-Months",
                 title="Rolling Beta to Benchmark",
                 lw=1.5,
                 fontname='Arial',
                 grayscale=False,
                 figsize=(10, 3),
                 ylabel=True,
                 subtitle=True,
                 savefig=None,
                 show=True):
    returns = pd.DataFrame(returns)
    beta1 = stats.rolling_greeks(returns, benchmark, window1)
    beta1 = beta1[beta1.columns.get_level_values(0)[0]]['beta']
    beta2= stats.rolling_greeks(returns, benchmark, window2)
    beta2 = beta2[beta2.columns.get_level_values(0)[0]]['beta']
    fig = _plot.rolling_beta(beta1, beta2,
                            window1_label=window1_label,
                            window2_label=window2_label,
                            title=title,
                            fontname=fontname,
                            grayscale=grayscale,
                            lw=lw,
                            figsize=figsize,
                            ylabel=ylabel,
                            subtitle=subtitle,
                            savefig=savefig,
                            show=show)
    if not show:
        return fig

def plt_rolling_volatility(returns, benchmark=None,
                       period=126, period_label="6-Months",
                       periods_per_year=252,
                       lw=1.5,
                       fontname='Arial',
                       grayscale=False,
                       figsize=(10, 3),
                       ylabel="Volatility",
                       subtitle=True,
                       savefig=None,
                       show=True):

    returns = stats.rolling_volatility(returns, period, periods_per_year)
    benchmark = stats.rolling_volatility(benchmark, period, periods_per_year)

    fig = _plot.rolling_stats(returns, benchmark,
                                   hline=returns.mean(),
                                   hlw=1.5,
                                   ylabel=ylabel,
                                   title='Rolling Volatility (%s)' % period_label,
                                   fontname=fontname,
                                   grayscale=grayscale,
                                   lw=lw,
                                   figsize=figsize,
                                   subtitle=subtitle,
                                   savefig=savefig, show=show)
    if not show:
        return fig

def plt_rolling_sharpe(returns, benchmark=None, rf=0.,
                       period=126, period_label="6-Months",
                       periods_per_year=252,
                       lw=1.25,
                       fontname='Arial',
                       grayscale=False,
                       figsize=(10, 3),
                       ylabel="Sharpe",
                       subtitle=True,
                       savefig=None,
                       show=True):

    returns = stats.rolling_sharpe(returns, rf, period, True, periods_per_year )

    if benchmark is not None:
        benchmark = stats.rolling_sharpe(benchmark, rf, period, True, periods_per_year)

    fig = _plot.rolling_stats(returns, benchmark,
                            hline=returns.mean(),
                            hlw=1.5,
                            ylabel=ylabel,
                            title='Rolling Sharpe (%s)' % period_label,
                            fontname=fontname,
                            grayscale=grayscale,
                            lw=lw,
                            figsize=figsize,
                            subtitle=subtitle,
                            savefig=savefig,
                            show=show)
    if not show:
        return fig


def plt_rolling_sortino(returns, benchmark=None, rf=0.,
                       period=126, period_label="6-Months",
                       periods_per_year=252,
                       lw=1.25,
                       fontname='Arial',
                       grayscale=False,
                       figsize=(10, 3),
                       ylabel="Sortino",
                       subtitle=True,
                       savefig=None,
                       show=True):

    returns = stats.rolling_sortino(returns, rf, period, True, periods_per_year )

    if benchmark is not None:
        benchmark = stats.rolling_sortino(benchmark, rf, period, True, periods_per_year)

    fig = _plot.rolling_stats(returns, benchmark,
                            hline=returns.mean(),
                            hlw=1.5,
                            ylabel=ylabel,
                            title='Rolling Sortino (%s)' % period_label,
                            fontname=fontname,
                            grayscale=grayscale,
                            lw=lw,
                            figsize=figsize,
                            subtitle=subtitle,
                            savefig=savefig,
                            show=show)
    if not show:
        return fig

def plt_drawdowns_periods(returns, periods=5,
                          lw=1.5,
                          log_scale=False,
                          fontname='Arial',
                          grayscale=False,
                          figsize=(10, 5),
                          ylabel=True,
                          subtitle=True,
                          savefig=None,
                          show=True):


    dd = utils.to_drawdown_series(returns)
    dddf = utils.drawdown_details(dd)
    longest_dd = dddf.sort_values(
        by='days', ascending=False, kind='mergesort')[:periods]

    returns = utils.to_rolling_returns(returns)

    fig = _plot.longest_drawdowns(returns, longest_dd, periods,
                                 lw=lw,
                                 log_scale=log_scale,
                                 fontname=fontname,
                                 grayscale=grayscale,
                                 figsize=figsize,
                                 ylabel=ylabel,
                                 subtitle=subtitle,
                                 savefig=savefig,
                                 show=show)
    if not show:
        return fig

def plt_drawdown(returns,
             title='Underwater Plot',
             grayscale=False,
             figsize=(10, 5),
             fontname='Arial',
             lw=1,
             log_scale=False,
             ylabel="Drawdown",
             subtitle=True,
             savefig=None,
             show=True):

    dd = utils.to_drawdown_series(returns)
    fig = _plot.timeseries(dd,
                          title=title,
                          hline=dd.mean(),
                          hlw=2,
                          hllabel="Average",
                          returns_label="Drawdown",
                          log_scale=log_scale,
                          fill=True,
                          lw=lw,
                          figsize=figsize,
                          ylabel=ylabel,
                          fontname=fontname,
                          grayscale=grayscale,
                          subtitle=subtitle,
                          savefig=savefig,
                          show=show)
    if not show:
        return fig

def plt_monthly_heatmap(returns,
             title='Monthly Returns',
             annot_size=10,
             figsize=(10, 5),
             cbar=True,
             square=False,
             eoy=False,
             grayscale=False,
             fontname='Arial',
             ylabel=False,
             savefig=None,
             show=True):

    returns = stats.monthly_returns(returns, eoy=eoy) * 100
    fig = _plot.heatmap(returns,
             title=title,
             grayscale=grayscale,
             figsize=figsize,
             fontname=fontname,
             ylabel=ylabel,
             savefig=savefig,
             show=show)
    
    if not show:
        return fig

def plt_distribution(returns,
                     fontname='Arial',
                     grayscale=False,
                     ylabel=True,
                     figsize=(10, 6),
                     subtitle=True,
                     savefig=None,
                     show=True):
  port = pd.DataFrame(returns.fillna(0))
  port.columns = ['Daily']

  port['Weekly'] = port['Daily'].resample(
      'W-MON').apply(stats.total_return)
  port['Weekly'].ffill(inplace=True)

  port['Monthly'] = port['Daily'].resample(
      'M').apply(stats.total_return)
  port['Monthly'].ffill(inplace=True)

  port['Quarterly'] = port['Daily'].resample(
      'Q').apply(stats.total_return)
  port['Quarterly'].ffill(inplace=True)

  port['Yearly'] = port['Daily'].resample(
      'A').apply(stats.total_return)
  port['Yearly'].ffill(inplace=True)

  fig = _plot.distribution(port,
                          fontname=fontname,
                          grayscale=grayscale,
                          figsize=figsize,
                          ylabel=ylabel,
                          subtitle=subtitle,
                          savefig=savefig,
                          show=show)
  if not show:
      return fig